"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest, egglib, pathlib

path = pathlib.Path(__file__).parent / '..' / 'data'

genepop1 = \
[[['AA8', '0405',' 0711', '0304', '0000', '0505'],
  ['AA9', '0405',' 0609', '0208', '0000', '0505'], 
  ['A10', '0205',' 0609', '0101', '0000', '0305'],
  ['A11', '0405',' 0606', '0002', '0000', '0504'],
  ['A12', '0202', '0609', '0105', '0000', '0507'],
  ['A13', '0505', '0909', '0100', '0000', '0505'],
  ['A14', '0202', '0609', '0207', '0000', '0503'],
  ['A15', '0405', '0609', '0101', '0000', '0505']],

 [['AF', '0000', '0000', '0000', '0000','0505'],
  ['AF', '0205', '0307', '0102', '0000','0505'],
  ['AF', '0202', '0609', '0202', '0000','0505'],
  ['AF', '0205', '0909', '0000', '0000','0505'],
  ['AF', '0205', '0307', '0202', '0000','0505'],
  ['AF', '0505', '0303', '0102', '0000','0505'],
  ['AF', '0205', '0700', '0000', '0000','0505'],
  ['AF', '0505', '0900', '0000', '0000','0405'],
  ['AF', '0205', '0600', '0000', '0000','0505'],
  ['AF', '0505', '0606', '0202', '0000','0505']],

 [['C45', '0505',' 0606', '0202', '0000','0505'],
  ['C45', '0505',' 0909', '0202', '0000','0505'],
  ['C45', '0505',' 0306', '0202', '0000','0505'],
  ['C45', '0505',' 0909', '0102', '0000','0405'],
  ['C45', '0205',' 0303', '0202', '0000','0505'],
  ['C45', '0205',' 0909', '0202', '0000','0405']]]

genepop2 = \
[[['RueDuQuai',  '250230', '564568', '110100'],
  ['RueDuQuai',  '252238', '568558', '100120'],
  ['RueDuQuai',  '254230', '564558', '090100'],
  ['RueDuQuai',  '250230', '564568', '110100'],
  ['RueDuQuai',  '252240', '568558', '100120'],
  ['RueDuQuai',  '254230', '564558', '090090']],

 [['Benitier',   '254230', '564558', '080100'],
  ['Benitier',   '000230', '564558', '090080'],
  ['Benitier',   '254230', '000000', '090100'],
  ['Benitier',   '254230', '564000', '090120']]]



class Genepop_test(unittest.TestCase):

    def check(self, data, n, aln):
        c = 0
        for i, pop in enumerate(data):
            for idv in pop:
                self.assertEqual(aln[c].name, idv[0]+'_1')
                self.assertListEqual(list(aln[c].sequence), [int(i.strip()[:n]) for i in idv[1:]])
                self.assertListEqual(list(aln[c].labels), ['pop{0}'.format(i+1), idv[0]])
                self.assertEqual(aln[c+1].name, idv[0]+'_2')
                self.assertListEqual(list(aln[c+1].sequence), [int(i.strip()[n:]) for i in idv[1:]])
                self.assertListEqual(list(aln[c+1].labels), ['pop{0}'.format(i+1), idv[0]])
                c += 2

    def test_genepop_E(self):
        with self.assertRaises(IOError):
            egglib.io.from_genepop('I don\'t exist!')
        with self.assertRaises(ValueError):
            egglib.io.from_genepop(path / 'genepop0.txt')

    def test_genepop_G(self):
        aln1 = egglib.io.from_genepop(path / 'genepop1.txt')
        self.assertIsInstance(aln1, egglib.Align)

        aln2 = egglib.io.from_genepop(path / 'genepop2.txt')
        self.assertIsInstance(aln2, egglib.Align)

        self.assertEqual(aln1.title, 'Microsat on Chiracus radioactivus, a pest species')
        self.assertListEqual(aln1.loci, ['Loc1', 'Loc2', 'Loc3', 'Y-linked', 'Loc4'])
        self.check(genepop1, 2, aln1)

        self.assertEqual(aln2.title, 'Title line: "Mosquito populations in southern France"')
        self.assertListEqual(aln2.loci, ['MicroSat 1', 'Microsat 2', 'Est-3'])
        self.check(genepop2, 3, aln2)

    def test_missing(self):
        aln1 = egglib.io.from_genepop(path / 'genepop1.txt')
        self.assertEqual(egglib.site_from_align(aln1, 0).num_missing, 2)
        self.assertEqual(egglib.site_from_align(aln1, 1).num_missing, 5)
        self.assertEqual(egglib.site_from_align(aln1, 2).num_missing, 12)
        self.assertEqual(egglib.site_from_align(aln1, 3).num_missing, 48)
        self.assertEqual(egglib.site_from_align(aln1, 4).num_missing, 0)



"""
    Copyright 2015-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from .. import eggwrapper as _eggwrapper
from .. import _interface
from .. import alphabets

def from_fasta(fname, alphabet, labels=False, label_marker='@', label_separator=',', cls=None):
    """
    Import sequences from a fasta file.
    Create a new instance of either :class:`.Align` or
    :class:`.Container` from data read from a fasta-formatted file.
    To process data from a fasta-formatted string, use
    :func:`.io.from_fasta_string`.

    :param source: name of a fasta-formatted sequence file.
    :param alphabet: an :class:`.Alphabet` instance defining the type
        of data. Only character alphabets are allowed (such as
        :py:obj:`.alphabets.DNA`, or :py:obj:`.alphabets.protein`).
    :param labels: import group labels. If so, they are not actually required to be
        present for each (or any) sequence. By default tags in sequence names
        considered to be part of the name and not as labels.
    :param label_marker: this option allows to change the character
        indicating the start of the labels.
    :param label_separator: this options allows to change character used
        to separate labels.
    :param cls: type that should be generated. Possible values are:
        :class:`!Align` (then, data must be aligned),
        :class:`!Container`, or ``None``. In the latter case, an
        :class:`!Align` is returned if data are found to be aligned or
        if the data set is empty, and otherwise a :class:`!Container` is
        returned.

    :return: A new :class:`.Container` or :class:`.Align` instance
        depending on the value of the *cls* option.
    """
    if not isinstance(alphabet._obj, _eggwrapper.CharAlphabet):
        raise ValueError('invalid alphabet for parsing fasta data: {0}'.format(alphabet.name))
    fasta_parser = _eggwrapper.FastaParser()
    fasta_parser.open_file(fname, alphabet._obj)
    return _from_fasta(fasta_parser, alphabet, labels, label_marker, label_separator, cls)

def from_fasta_string(string, alphabet, labels=False, label_marker='@', label_separator=',', cls=None):
    """
    Import sequences from a fasta-formatted string. Identical
    to :func:`.io.from_fasta` but directly takes an fasta-formatted string as first argument.
    """
    if not isinstance(alphabet._obj, _eggwrapper.CharAlphabet):
        raise ValueError('invalid alphabet for parsing fasta data: {0}'.format(alphabet.name))
    fasta_parser = _eggwrapper.FastaParser()
    fasta_parser.set_string(string, alphabet._obj)
    return _from_fasta(fasta_parser, alphabet, labels, label_marker, label_separator, cls)

def _from_fasta(fasta_parser, alphabet, labels, label_marker, label_separator, cls):
    obj = _eggwrapper.DataHolder(False)
    fasta_parser.read_all(labels, obj, label_marker, label_separator)
    if cls is _interface.Align or cls is None:
        ns = set([obj.get_nsit_sample(i) for i in range(obj.get_nsam())])
        if len(ns) == 0:
            ns = 0
            if cls is None: cls = _interface.Align
        elif len(ns) == 1:
            ns = ns.pop()
            if cls is None: cls = _interface.Align
        else:
            if cls is _interface.Align:
                raise ValueError('cannot create `Align`: lengths of sequences do not match')
            cls = _interface.Container
    elif cls is not _interface.Container:
        raise ValueError('invalid value provided for `cls`')

    if cls is _interface.Container: return _interface.Container._create_from_data_holder(obj, alphabet)
    else: return _interface.Align._create_from_data_holder(obj, alphabet)

class fasta_iter(object):
    """
    Iterative sequence-by-sequence fasta parser.
    
    :param fname: name of a fasta-formatted file.
    :param alphabet: an :class:`.Alphabet` instance defining the type
        of data. Only character alphabets are allowed (such as
        :py:obj:`.alphabets.DNA` and :py:obj:`.alphabets.protein`).
    :param labels: import group labels from sequence names
        (by default, they are considered as part of the name).

    This function can be
    used in an iteration as shown below:

    .. code-block:: python

        >>> for item in egglib.io.fasta_iter(fname):
        ...     ...

    The ``with`` statement is also supported, which
        ensures that the input file is properly closed whenever the
        ``with`` statement completes:

    .. code-block:: python

       >>> with egglib.io.fasta_iter(fname) as f:
       ...     for item in f:
       ...         ...

    Each iteration yields a :class:`.SampleView` instance (which is
    valid only during the iteration round, see the warning below).

    .. warning::
        The aim of this iterator is to iterate over large fasta files
        without actually storing all data in memory at the same time.
        The :class:`.SampleView` instance provided at each iteration is a proxy to
        a local :class:`.Container` instance that is recycled at each
        iteration step. The iteration variable should be used immediately
        and never stored as this. If one wants to sequence data, they should
        copy them immediately (typically using the
        :meth:`~.Container.add_sample` method of a separate
        :class:`.Container` instance).
    """

    def __init__(self, fname, alphabet, labels=False):
        self._parser = _eggwrapper.FastaParser() # define it before, otherwise the exit code will break in case of error
        if not isinstance(alphabet._obj, _eggwrapper.CharAlphabet):
            raise ValueError('invalid alphabet for parsing fasta data: {0}'.format(alphabet.name))
        self._parser.open_file(fname, alphabet._obj)
        self._cont = _interface.Container(alphabet)
        self._labels = labels

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._parser.close()

    def __del__(self):
        self._parser.close()

    def __iter__(self):
        return self

    def __next__(self):
        if not self._parser.good(): raise StopIteration
        self._cont.reset()
        self._parser.read_sequence(self._labels, self._cont._obj)
        self._cont._ns = self._cont._obj.get_nsam()
        return _interface.SampleView(self._cont, 0)



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, unittest

class SiteDiversity_test(unittest.TestCase):
    def test_return_value(self): # not all cases covered
        alph = egglib.alphabets.Alphabet('int', [0, 1], [-1])
        site = egglib.Site()
        frq = egglib.Freq()
        sd = egglib._eggwrapper.SiteDiversity()

        site.from_list([-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1], alphabet=alph)
        frq.from_site(site)
        self.assertEqual(sd.process(frq._obj), 1)
        self.assertEqual(sd.ns(), 0)

        site.from_list([0,1,0,0,1,1,0,0,0,1,1,1,0,0,0,1,1,1,1,0,1,1,0,1], alphabet=alph)
        frq.from_site(site)
        self.assertEqual(sd.process(frq._obj), 1+2+4+512+1024)
        self.assertEqual(sd.ns(), 24)

        site.from_list([0,1,0,0,-1,-1,0,-1,-1,1,1,1,0,0,-1,1,1,1,1,0,-1,1,0,1], alphabet=alph)
        frq.from_site(site)
        self.assertEqual(sd.process(frq._obj), 1+2+4+512+1024)
        self.assertEqual(sd.ns(), 18)

        site.from_list([0,-1,0,0,-1,-1,0,-1,-1,-1,-1,-1,0,0,-1,-1,-1,-1,-1,0,-1,-1,0,-1], alphabet=alph)
        frq.from_site(site)
        self.assertEqual(sd.process(frq._obj), 1+2+4+512)
        self.assertEqual(sd.ns(), 8)



    def test_Fis(self):
        # simulation model
        NR = 200
        struct = [55, 32, 65, 58]
        sim = egglib.coalesce.Simulator(len(struct), num_indiv=struct,
                                        theta=5, num_sites=1, mut_model='SMM')
        sim.params['events'].add(T=0.2, cat='merge', src=1, dst=0)
        sim.params['events'].add(T=0.2, cat='merge', src=2, dst=0)
        sim.params['events'].add(T=0.2, cat='merge', src=3, dst=0)

        # stats calculator
        cs = egglib.stats.ComputeStats(struct=sim.params.mk_structure(), multi_hits=True)
        cs.add_stats('Fis', 'Ho', 'He')
        cs_tot = egglib.stats.ComputeStats(struct=sim.params.mk_structure(), multi_hits=True, multi=True)
        cs_tot.add_stats('Fis', 'Ho', 'He')

        # simulate datasets and compute Fis with EggLib
        alph = egglib.Alphabet('range', (-999999, None), (-1000000, -999999))
        He_tot = 0.0
        Ho_tot = 0.0
        for aln in sim.iter_simul(NR):
            aln = egglib.Align.create(list(aln), alphabet=alph)
            aln.random_missing(0.05, missing=-1000000)
            stats = cs.process_align(aln, max_missing=1)
            cs_tot.process_align(aln, max_missing=1)
            Fis = stats['Fis']
            Ho = stats['Ho']
            He = stats['He']

            # compute Fis manually
            alleles = {}
            heter = 0
            nind = 0
            for indiv in range(sum(struct)):
                a = aln[indiv*2].sequence[0]
                b = aln[indiv*2+1].sequence[0]
                if alph.get_code(a) >= 0:
                    if a not in alleles: alleles[a] = 0
                    alleles[a] += 1
                if alph.get_code(b) >= 0:
                    if b not in alleles: alleles[b] = 0
                    alleles[b] += 1
                if alph.get_code(a) >= 0 and alph.get_code(b) >= 0:
                    nind += 1
                    if a != b:
                        heter += 1

            He_check = 1.0
            nseff = sum(alleles.values())
            if nseff < 2: continue
            for a in alleles:
                He_check -= (alleles[a]/nseff) ** 2
            He_check *= nseff / (nseff - 1)
            Ho_check = heter / nind
            Fis_check = 1 - Ho_check / He_check
            He_tot += He_check
            Ho_tot += Ho_check

            self.assertAlmostEqual(Fis, Fis_check, places=6)
            self.assertAlmostEqual(He, He_check, places=6)
            self.assertAlmostEqual(Ho, Ho_check, places=6)

        stats = cs_tot.results()
        Fis_tot = 1 - Ho_tot / He_tot
        self.assertAlmostEqual(stats['He'], He_tot/NR, places=6)
        self.assertAlmostEqual(stats['Ho'], Ho_tot/NR, places=6)
        self.assertAlmostEqual(stats['Fis'], Fis_tot, places=6)



"""
    Copyright 2008-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys, re, functools
from .. import _interface
from .. import alphabets

_iupac = {
    '?': set('ACGT-'),
    '-': set('-'),
    'A': set('A'),
    'C': set('C'),
    'G': set('G'),
    'T': set('T'),
    'M': set('AC'),
    'R': set('AG'),
    'W': set('AT'),
    'S': set('CG'),
    'Y': set('CT'),
    'K': set('GT'),
    'B': set('CGT'),
    'D': set('AGT'),
    'H': set('ACT'),
    'V': set('ACG'),
    'N': set('ACGT')
}

for _i in list(_iupac.keys()):
    if _i != '?' and _i != '-': # it would be no harm but awkward
        _iupac[_i.lower()] = _iupac[_i]

_iupac_set = set(_iupac.keys())
_bases_set = _iupac_set - set('-?')

_trans = dict(zip('ACGT-?', 'TGCA-?'))

_a = []
_b = []
for _i, _values in _iupac.items():
    if _i.islower(): continue
    values = set([_trans[_j] for _j in _values])
    for _j, _k in _iupac.items():
        if _k == values:
            _a.append(_i)
            _b.append(_j)
            if _i.lower() != _i:
                _a.append(_i.lower())
                _b.append(_j.lower())
            break
    else: raise RuntimeError('internal error')
_rc_conv =  str.maketrans(''.join(_a), ''.join(_b))

_regex_map = {}
for _i, _j in _iupac.items():
    _map = set()
    for _m, _n in _iupac.items():
        if _m.isupper():
            for _k in _n:
                 if _k not in _j: break
            else:
                _map.add(_m)
    _regex_map[_i] = _map.pop() if len(_map) == 1 else '[' + ''.join(_map) + ']'

_gap_codes = {
    'DNA': [alphabets.DNA.get_code('-')],
    'protein': [alphabets.protein.get_code('-')],
    'codons': [alphabets.codons.get_code(allele) for allele in alphabets.codons.get_alleles()[1] if '-' in allele]
}

def rc(seq):
    """
    Reverse-complement a DNA sequence.

    :param seq: input nucleotide sequence, as a :class:`str`, or a :class:`.SequenceView`.

    :return: The reverse-complemented sequence (see details below).

    The case of the provided sequence is preserved. Ambiguity characters
    are complemented according to the table shown :ref:`here <iupac-nomenclature>`.
    Invalid characters characters raise a :exc:`ValueError`.
    """
    if isinstance(seq, _interface.SequenceView): seq = seq.string()
    if not _iupac_set.issuperset(seq): raise ValueError('invalid character in DNA sequence')
    return seq[::-1].translate(_rc_conv) 

def compare(seq1, seq2):
    """
    Compare two sequences. The comparison supports ambiguity characters
    (see :ref:`here <iupac-nomenclature>`) such that partially overlapping
    ambiguity sets characters are not treated as different. For example,
    ``A`` and ``M`` are not treated as different, nor are ``M`` and
    ``R``. Only IUPAC characters may be supplied.

    :param seq1: a DNA sequence (as a :class:`str` or  a :class:`.SequenceView`).
    :param seq2: another DNA sequence (idem).

    :return: ``True`` if sequences have the same length and they either
        are identical or differ only by overlapping IUPAC characters,
        ``False`` otherwise.
    """
    
    if isinstance(seq1, _interface.SequenceView): seq1 = seq1.string()
    if isinstance(seq2, _interface.SequenceView): seq2 = seq2.string() 
    
    if len(seq1) != len(seq2):
        return False

    try:
        for i, j in zip(seq1, seq2):
                if _iupac[i].isdisjoint(_iupac[j]):
                    return False
    except KeyError as e:
        raise ValueError('unknown DNA base: {0}'.format(e.args[0]))

    return True

def regex(query, both_strands=False):
    """
    Turn a DNA sequence into a regular expression. The input sequence
    should contain IUPAC characters only (see
    :ref:`here <iupac-nomenclature>`). Ambiguity characters will be teated as
    follows: an ambiguity will match either (1) itself, (2) one of the
    characters it defines, or (3) one of the ambiguity characters that
    define a subset of the characters it defines. For example, a ``M``
    in the query will match an ``A``, a ``C`` or a ``M``, and a ``D``
    will match all the following: ``A``, ``G``, ``T``, ``R``, ``W``,
    ``K``, and ``D``. The fully degenerated ``N`` matches all four bases
    and all intermediate ambiguity characters and, finally, ``?``
    matches all allowed characters (including alignment gaps). Note ``-``
    only matches itself.

    :param query: a :class:`str` containing IUPAC characters or a :class:`.SequenceView`.

    :param both_strands: look for the query on both forward and reverse
        strands (by default, only on forward strand).

    :return: A regular expression as a  :class:`!str` expanding ambiguity
        characters to all compatible characters.

    Result of this function can be used with the module :mod:`re` to
    locate occurrences of a motif or the position of a sequence as in:

    .. code:: python

        >>> regex = egglib.tools.regex(query)
        >>> for hit in re.finditer(regex, subject):
        ...     print(hit.start(), hit.end(), hit.group(0))

    The returned regular expression includes upper-case characters,
    regardless of the case of input characters. To perform a
    case-insensitive search, use the :data:`re.IGNORECASE` flag of the
    :mod:`!re` module in regular expression searches (as in
    ``re.search(egglib.tools.regex(query), subject, re.IGNORECASE)``.
    Note that the regular expression is contained into a group if
    *both_strands* is ``False``, and two groups otherwise (one for each
    strand).
    """
    
    if isinstance(query, _interface.SequenceView): query = query.string()
    
    chars = list(query)
    for i, v in enumerate(chars):
        if v not in _regex_map: raise ValueError('invalid character in DNA sequence')
        chars[i] = _regex_map[v]
    if both_strands: return '({0})|{1}'.format(''.join(chars), regex(rc(query), False))
    else: return '(' + ''.join(chars) + ')'

def motif_iter(subject, query, mismatches=0, both_strands=False, case_independent=True, only_base=True):
    """
    Iterate over hits of a given query.
    Return an iterator over hits of the provided query in the subject
    sequence. The query should only contain bases (including IUPAC
    ambiguity characters as listed in :ref:`here <iupac-nomenclature>`, but
    excluding ``-`` and ``?``). Ambiguity characters are treated as
    described in :func:`.regex` (the bottom line is that ambiguities in
    the query can only match identical or less degenerate ambiguities in
    the subject). Mismatches are allowed, and the iterator will first
    yield identical hits (if they exist), then hits with one mismatch,
    then hits with two mismatches, and so on.

    :param subject: a DNA sequence as :class:`str` or a :class:`.SequenceView`. The sequence may
        contain ambiguity characters. In principle, the subject sequence
        should contain IUPAC characters only, but this is not strictly
        enforced (non-IUPAC characters will always be treated as
        different of IUPAC characters).

    :param query: a DNA sequence as :class:`!str` or a :class:`!SequenceView`, also containing
        IUPAC characters only.

    :param mismatches: maximum number of mismatches.

    :param both_strands: look for the query on both forward and reverse
        strands (by default, only on forward strand).

    :param case_independent: if ``True``, perform case-independent
        searches.

    :param only_base: if ``True``, never create a hit if the putative
        hit sequence contains an alignment gap (``-``) or a ``?``
        character, irrespective of the number of allowed mismatches.

    :return: An iterator returning, for each hit, a
        ``(start, stop, strand, num)`` tuple where ``start`` and ``stop``
        are the position of the hit such that ``subject[start:stop]``
        returns the hit sequence, ``strand`` is the strand as ``+`` or
        ``-``, and ``num`` is the number of mismatches of the hit.

    .. note::

        If a given query has a hit on both strands at the same position
        (this only happens if the sequence is a motif equal to its
        reverse-complement like ``AATCGATT``), it is guaranteed that the
        only one hit will be returned for a given position (no twice on
        each strand). However, the value of the ``strand`` flag is not
        defined.

    .. warning::

        This function is designed to be efficient only if the number of
        mismatches is small (only a few). A combination of a large
        number of mismatches (more than 2 or 3) and a large query
        sequence (as early as 10 or 20), will take significant time to
        complete. More complex problems require the use of genuine
        pairwise ocal alignment tools.
    """

    if isinstance(subject, _interface.SequenceView): subject = subject.string()
    if isinstance(query, _interface.SequenceView): query = query.string()
    
    # get query length
    ln = len(query)
    if ln < 1: raise ValueError('query must not be empty')
    if not _bases_set.issuperset(query): raise ValueError('invalid character in DNA sequence')
    if mismatches > ln: raise ValueError('too many mismatches allowed')

    # initialize table of positions of mismatches
    # if n>1, mismatch indexes are always increasing (first value for n=3 is [0,1,2])
    n = 1 # pos cannot be incremented if 0
    pos = [-1] # the last (and only) index will be incremented once before use

    # current query (initially without mismatches)
    cur = list(query)

    # cache to avoid returning several times the same hit
    cache = set()

    # process all possibilities (no mismatches, then 1, then 2, etc.)
    while True:

        # perform search (will be done at least once)
        for mo in re.finditer(regex(''.join(cur), both_strands), subject, re.I if case_independent else 0):
            if mo.start() not in cache:
                yield mo.start(), mo.end(), '+-'[mo.lastindex-1], 0 if pos[0]==-1 else n
                cache.add(mo.start())

        # increment mismatch index (initialized to -1)
        i = n -1 # current mismatch index
        pos[i] += 1 # increment last index
        stop = ln # stop value (maximum pos for n=3 and ln=10 is [7,8,9])

        # increment previous index if end is reached (and so one)
        while pos[i] == stop:

            # all mismatches processed, add a new mismatch
            if i == 0:
                n += 1
                pos = list(range(n))
                break

            # increment previous index
            else:
                i -= 1
                stop -= 1
                pos[i] += 1
                pos[i+1:] = range(pos[i]+1, pos[i]+n-i)

        # quit if already too many mismatches
        if n > mismatches: break

        # create a list from the query with allowed mismatches (at least 1)
        cur = list(query)
        for i, v in enumerate(pos):
            cur[v] = 'N' if only_base else '?'

def _get_gaps(gaps, alph):
    # helper to validate gap option
    if gaps is None:
        gap_codes = _gap_codes.get(alph.name, None)
        if gap_codes is None: raise ValueError('`gap` argument is required for non-standard alphabets')
    else:
        if not isinstance(gaps, (list, tuple)):
            raise TypeError('invalid type for gaps')
        if len(gaps) == 0:
            raise ValueError('at least one gap value is expected')
        try:
            gap_codes = list(map(alph.get_code, gaps))
        except ValueError:
            raise ValueError('invalid gap value for the input alphabet')
    return gap_codes

def ungap_all(obj, gaps=None):
    """
    Remove all gaps from an object.
    Generate a :class:`.Container` instance containing all sequences of
    the provided object excluding any gaps.

    :param obj: input object as an :class:`.Align` or a :class:`.Container`.

    :param gaps: the list of allelic value(s) representing gaps. The
        argument must be a :class:`list` or :class:`tuple` containing
        one gap allele value, or more than one if appropriate. All
        values must be valid alleles for the input alphabet. By
        default, use the ``-`` character for the DNA and protein
        alphabets, and all triplets containing at least one alignment
        gaps for the codon alphabet. There is no default for other
        alphabets.

    :return: A :class:`!Container` instance.
    """
    if not isinstance(obj, (_interface.Align, _interface.Container)): raise TypeError('expect an Align or a Container instance')
    gap_codes = _get_gaps(gaps, obj._alphabet)
    ls = obj.ls
    result = _interface.Container(alphabet=obj._alphabet)
    for i in range(obj.ns):
        result.add_sample(
            obj.get_name(i),
            [obj._alphabet.get_value(obj._obj.get_sample(i, j)) for j in range(ls) if obj._obj.get_sample(i, j) not in gap_codes],
            obj.get_sample(i).labels)
    return result

def ungap(align, freq, gaps=None):
    """
    Remove sites with too many gaps.
    Generate a new :class:`.Align` instance containing all sequences of
    the provided alignment but with only those sites for which the
    frequency of alignment gaps is less than or equal to *freq*.

    :param align: input alignment as an :class:`!Align`.

    :param freq: maximum gap frequency in a site (if there are more
        gaps, the site is not included in the returned :class:`.Align`
        instance). This value is a
        relative frequency (included in the [0, 1] range).

    :param gaps: the list of allelic value(s) representing gaps. The
        argument must be a :class:`list` or :class:`tuple` containing
        one gap allele value, or more than one if appropriate. All
        values must be valid alleles for the alignment alphabet. By
        default, use the ``-`` character for the DNA and protein
        alphabets, and all triplets containing at least one alignment
        gaps for the codon alphabet. There is no default for other
        alphabets.

    :return: A new :class:`!Align` instance.
    """

    # process arguments
    if not isinstance(align, _interface.Align): raise TypeError('expect an Align instance')
    gap_codes = _get_gaps(gaps, align._alphabet)

    # get statistics
    ls = align.ls
    if freq < 0.0 or freq > 1.0: raise ValueError('`freq` value is out of range')
    ns = align.ns
    if ns == 0: raise ValueError('not enough sequences in alignment')

    # collect retained positions
    good = []
    for i in range(ls):
        gaps = 0
        for j in range(ns):
            if align._obj.get_sample(j, i) in gap_codes: gaps += 1
        if gaps/ns <= freq:
            good.append(i)

    # return requested object
    return align.extract(good)



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, unittest, pathlib

ALPH = egglib.alphabets.Alphabet('range', [0, None], [-1, 0], case_insensitive=False)

class LD_test(unittest.TestCase):
    def test_pairwise_LD_T(self):
        sites = [egglib.site_from_list(i, ALPH) for i in 
                 [  [0,0,0,1,1,1,2,1,1,1],
                    [0,1,1,-1,0,0,0, 0,0,0],
                    [0,0,1,2, 2,1,-1,2,-1,2]]  ]
        ld=egglib.stats.pairwise_LD(sites[0], sites[1])
        self.assertIsInstance(ld, dict)

        # additional tests addressing the structure
        coal = egglib.coalesce.Simulator(num_pop=2, num_chrom=[25, 25], theta=5, migr=0.5)
        alnA = coal.simul()
        structA = egglib.struct_from_labels(alnA, lvl_pop=0)
        structB = egglib.struct_from_dict({None: {None: structA.as_dict()[0][None]['0']}}, None)
        alnB = egglib.Align.create([(sam.name, sam.sequence) for sam in alnA.group_mapping()['0']], alphabet=alnA.alphabet)

        stats1 = egglib.stats.matrix_LD(alnA, ['d', 'D', 'Dp', 'r'], struct=structA)
        stats2 = egglib.stats.matrix_LD(alnA, ['d', 'D', 'Dp', 'r'], struct=None)
        self.assertEqual(stats1, stats2)

        stats3 = egglib.stats.matrix_LD(alnA, ['d', 'D', 'Dp', 'r'], struct=structB)
        stats4 = egglib.stats.matrix_LD(alnB, ['d', 'D', 'Dp', 'r'], struct=structB) # should work also with None but order can be screwed and then r, D (Dp?) can be inverted
        self.assertEqual(stats3, stats4)
        self.assertNotEqual(stats1, stats3)

        siteA1 = egglib.site_from_align(alnA, 0)
        siteA2 = egglib.site_from_align(alnA, 1)
        siteAL = egglib.site_from_align(alnA, alnA.ls - 1)

        siteB1 = egglib.site_from_align(alnB, 0)
        siteB2 = egglib.site_from_align(alnB, 1)
        siteBL = egglib.site_from_align(alnB, alnB.ls - 1)

        statsA = egglib.stats.pairwise_LD(siteA1, siteA2), egglib.stats.pairwise_LD(siteA1, siteAL)
        statsAsA = egglib.stats.pairwise_LD(siteA1, siteA2, struct=structA), egglib.stats.pairwise_LD(siteA1, siteAL, struct=structA)
        statsAsB = egglib.stats.pairwise_LD(siteA1, siteA2, struct=structB), egglib.stats.pairwise_LD(siteA1, siteAL, struct=structB)
        statsB = egglib.stats.pairwise_LD(siteB1, siteB2), egglib.stats.pairwise_LD(siteB1, siteBL)

        for k in 'D', 'Dp', 'r':
            if statsA[0][k] is not None: statsA[0][k] = abs(statsA[0][k])
            if statsA[1][k] is not None: statsA[1][k] = abs(statsA[1][k])
            if statsAsA[0][k] is not None: statsAsA[0][k] = abs(statsAsA[0][k])
            if statsAsA[1][k] is not None: statsAsA[1][k] = abs(statsAsA[1][k])
            if statsAsB[0][k] is not None: statsAsB[0][k] = abs(statsAsB[0][k])
            if statsAsB[1][k] is not None: statsAsB[1][k] = abs(statsAsB[1][k])
            if statsB[0][k] is not None: statsB[0][k] = abs(statsB[0][k])
            if statsB[1][k] is not None: statsB[1][k] = abs(statsB[1][k])

        self.assertListEqual(sorted(statsA[0]), sorted(statsA[1]))
        self.assertListEqual(sorted(statsA[0]), sorted(statsAsA[0]))
        self.assertListEqual(sorted(statsA[0]), sorted(statsAsA[1]))
        self.assertListEqual(sorted(statsA[0]), sorted(statsAsB[0]))
        self.assertListEqual(sorted(statsA[0]), sorted(statsAsB[1]))
        self.assertListEqual(sorted(statsA[0]), sorted(statsB[0]))
        self.assertListEqual(sorted(statsA[0]), sorted(statsB[1]))

        for k in statsA[0]:
            self.assertAlmostEqual(statsA[0][k], statsAsA[0][k])
            self.assertAlmostEqual(statsA[1][k], statsAsA[1][k])
            self.assertAlmostEqual(statsB[0][k], statsAsB[0][k])
            self.assertAlmostEqual(statsB[1][k], statsAsB[1][k])

        alph1 = egglib.alphabets.Alphabet('int', [1,2], [])
        alph2 = egglib.alphabets.Alphabet('int', [111,112,122,222], [])
        site1 = egglib.site_from_list([1,1,1,1,2,1,1,2,2,2,2,2,1,1,2], alphabet=alph1)
        site2 = egglib.site_from_list([1,1,1,1,1,1,1,1,2,2,2,2,1,2,2], alphabet=alph1)
        site1g = egglib.site_from_list([111,112,122,222,112], alphabet=alph2)
        site2g = egglib.site_from_list([111,111,112,222,122], alphabet=alph2)
        struct = egglib.struct_from_dict({None: {None: {'a': (0,1,2), 'b': (3,4,5), 'c': (6,7,8), 'd': (9,10,11), 'e': (12,13,14)}}}, None)

        stats1 = egglib.stats.pairwise_LD(site1, site2, struct=struct)
        stats2 = egglib.stats.pairwise_LD(site1g, site2g)
        stats3 = egglib.stats.pairwise_LD(site1, site2)

        for k in 'D', 'Dp', 'r':
            stats1[k] = abs(stats1[k])
            stats2[k] = abs(stats2[k])
            stats3[k] = abs(stats3[k])
        self.assertListEqual(sorted(stats1), sorted(stats2))
        self.assertListEqual(sorted(stats1), sorted(stats3))
        diff = 0
        for k in stats1:
            self.assertAlmostEqual(stats1[k], stats2[k])
            if abs(stats1[k] - stats3[k]) > 0.000001: diff += 1
        self.assertNotEqual(diff, 0, msg="expect at least one difference between stats")

    def test_pairwise_LD_E(self):
        sites = [egglib.site_from_list(i, ALPH)for i in 
                 [  [ 0, 0, 0, 1, 1, 1, 2, 1, 1, 1],
                    [ 0, 1, 1,-1, 0, 0, 0, 0],      ]   ]
        with self.assertRaises(ValueError):
            egglib.stats.pairwise_LD(sites[0], sites[1])

    def test_matrix_LD_T(self):
        p3 = egglib._eggwrapper.Params(1)
        p3.set_n1(0, 60)
        p3.set_mutmodel(p3.KAM)
        p3.set_K(2)
        p3.set_R(0.1)
        p3.set_theta(5.0)
        p3.set_L(1000)
        p3.autoSitePos()
        p3.validate()
        coal = egglib._eggwrapper.Coalesce()
        coal.simul(p3, True)
        simdata5 = coal.data()
        aln = egglib.Align(egglib.alphabets.DNA)
        for i in range(simdata5.get_nsam()):
            n = 'sample_{0}'.format(str(i).rjust(3, '0'))
            s = ''.join(['ACGT'[simdata5.get_sample(i, j)] for j in range(simdata5.get_nsit_all())])
            aln.add_sample(n, s)
        MLD_D=egglib.stats.matrix_LD(aln, 'D')
        self.assertIsInstance(MLD_D, tuple)
        MLD_all=egglib.stats.matrix_LD(aln, ['d', 'Dp', 'rsq'])
        self.assertIsInstance(MLD_all, tuple)
        
    def test_matrix_LD_E(self):
        p3 = egglib._eggwrapper.Params(1)
        p3.set_n1(0, 60)
        p3.set_mutmodel(p3.KAM)
        p3.set_K(2)
        p3.set_R(0.1)
        p3.set_theta(5.0)
        p3.set_L(1000)
        p3.autoSitePos()
        p3.validate()
        coal = egglib._eggwrapper.Coalesce()
        coal.simul(p3, True)
        simdata5 = coal.data()
        aln = egglib.Align(egglib.alphabets.DNA)
        for i in range(simdata5.get_nsam()):
            n = 'sample_{0}'.format(str(i).rjust(3, '0'))
            s = ''.join(['ACGT'[simdata5.get_sample(i, j)] for j in range(simdata5.get_nsit_all())])
            aln.add_sample(n, s)

        with self.assertRaises(ValueError):
            egglib.stats.matrix_LD(aln, 'D', min_n=1)
        with self.assertRaises(ValueError):
            egglib.stats.matrix_LD(aln, 'D', max_maj=1.5)
        with self.assertRaises(ValueError):
            egglib.stats.matrix_LD(aln, 'D', positions=[1001,1002,1003])
        with self.assertRaises(ValueError):
            egglib.stats.matrix_LD(aln, ['FAIL', 'Dp', 'rsq'])



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, unittest, collections

class ReadingFrame_test(unittest.TestCase):
    def test_readingframe_T(self):
        RF=egglib.tools.ReadingFrame()
        self.assertEqual(str(type(RF)),"<class 'egglib.tools._reading_frame.ReadingFrame'>")

    def test_process_T(self):
        RF=egglib.tools.ReadingFrame()	
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        self.assertEqual(str(type(RF)),"<class 'egglib.tools._reading_frame.ReadingFrame'>")
        self.assertEqual(RF.num_exons,4)

    def test_process_E(self):
        RF=egglib.tools.ReadingFrame()
        with self.assertRaises(ValueError):
            RF.process([(974, 1222, 17)])
        with self.assertRaises(ValueError):
            RF.process([(-10, 24)])
        with self.assertRaises(ValueError):
            RF.process([(24, 24)])
        with self.assertRaises(ValueError):
            RF.process([(24, 4)])
    
    def test_num_needed_bases_T(self):
        RF=egglib.tools.ReadingFrame()
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        self.assertEqual(RF.num_needed_bases, 1222)

    def test_num_tot_bases_T(self):
        RF=egglib.tools.ReadingFrame()	
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        self.assertEqual(RF.num_tot_bases,1212)

    def test_num_exon_bases_T(self):
        RF=egglib.tools.ReadingFrame()	
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        self.assertEqual(RF.num_exon_bases,411)

    def test_num_exons_T(self):
        RF=egglib.tools.ReadingFrame()	
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        self.assertEqual(RF.num_exons,4)

    def test_num_full_codons_T(self):
        RF=egglib.tools.ReadingFrame()
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        self.assertEqual(RF.num_codons, 136)
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)], keep_truncated=True)
        self.assertEqual(RF.num_codons, 139)

    def test_exon_index_T(self):
        RF=egglib.tools.ReadingFrame()	
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        self.assertEqual(RF.exon_index(1150),3)

    def test_codon_index_T(self):
        RF=egglib.tools.ReadingFrame()
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        self.assertEqual(RF.codon_index(1150), 112)

    def test_codon_position_T(self):
        RF=egglib.tools.ReadingFrame()	
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        self.assertEqual(RF.codon_position(1150),1)

    def test_codon_bases_T(self):
        RF=egglib.tools.ReadingFrame()	
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        self.assertEqual(RF.codon_bases(1),(13, 14, 15))

    def test_codon_bases_E(self):
        RF=egglib.tools.ReadingFrame()	
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        self.assertEqual(RF.codon_bases(4000),None)

    def test_iter_exon_bounds_T(self):
        RF=egglib.tools.ReadingFrame()	
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        l_ex=[10,55,255,974,24,144,315,1222]
        i=0
        self.assertIsInstance(RF.iter_exon_bounds(), collections.abc.Iterable)
        for a, b in RF.iter_exon_bounds(): 
            self.assertIn(a, l_ex)
            self.assertIn(a, l_ex)
            i=+1

    def test_iter_codons_T(self):
        RF=egglib.tools.ReadingFrame()	
        RF.process([(10, 24), (55, 144), (255, 315), (974, 1222, 3)])
        self.assertIsInstance(RF.iter_codons(), collections.abc.Iterable)



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import os, egglib, sys, unittest, random, re, gc
import collections

TREE_STRING = '(((VvLYK2:0.09020761,VvLYK3:0.13588809)1000:0.0759631,(PtLYK3:0.17773083,((FvLYK2:0.1079844,(PpLYK3:0.0532316,MdLYK3:0.11516319)1000:0.02510313)1000:0.07250312,((((CacLYK3:0.06016158,GmLYK3:0.04988316)998:0.04732511,LjLYS7:0.08887798)598:0.01845951,MtLYK8:0.17283636)1000:0.15093267,(CasLYK2:0.00284497,CasLYK3:0.00922375)1000:0.16983704)341:0.01703062)492:0.02813366)999:0.06833724)999:0.03856292,(VvLYK1:0.14879144,((PtLYK1:0.03931642,PtLYK2:0.06215206)1000:0.13921412,((CasLYK1:0.16961626,(AtCERK1:0.49326411,(FvLYK1:0.1453431,((PpLYK1:0.06109585,PpLYK2:0.07592721)933:0.02843438,(MdLYK1:0.0383793,MdLYK2:0.05889107)1000:0.04350185)924:0.02731021)999:0.05725737)294:0.00688996)430:0.02167592,(((CacLYK1:0.08167788,(GmNFR1a:0.03699912,GmNFR1b:0.02880309)677:0.01953081)998:0.0365518,(LjNFR1a:0.07999007,(MtLYK2:0.05186283,(MtLYK3:0.07978386,PsSYM37:0.07493097)822:0.014712)1000:0.05694231)832:0.0215714)1000:0.18063742,((CecLYK1:0.1680513,(((LjNFR1b:0.13030045,MtLYK1:0.15554318)1000:0.07303159,(LjNFR1c:0.13766267,MtLYK6:0.16025718)982:0.03580008)990:0.02671666,((GmLYK2:0.04933436,CacLYK4:0.05585385)1000:0.06029768,MtLYK7:0.1292539)997:0.04581387)1000:0.05564467)998:0.03954658,(CecLYK2:0.07693512,((CacLYK2:0.07392427,GmLYK2b:0.05439578)1000:0.05710725,(LjLYS6:0.06218823,MtLYK9:0.12383649)905:0.02507435)1000:0.08050891)794:0.01706889)820:0.01986008)999:0.06770856)577:0.02895946)942:0.04476039)999:0.03856292);'
TREE_STRING_R ='((((((((((LjNFR1b:0.13030045,MtLYK1:0.15554318)1000:0.07303159,(LjNFR1c:0.13766267,MtLYK6:0.16025718)982:0.03580008)990:0.02671666,((GmLYK2:0.04933436,CacLYK4:0.05585385)1000:0.06029768,MtLYK7:0.1292539)997:0.04581387)1000:0.05564467,CecLYK1:0.1680513)998:0.03954658,(((CacLYK2:0.07392427,GmLYK2b:0.05439578)1000:0.05710725,(LjLYS6:0.06218823,MtLYK9:0.12383649)905:0.02507435)1000:0.08050891,CecLYK2:0.07693512)794:0.01706889)820:0.01986008,((((MtLYK3:0.07978386,PsSYM37:0.07493097)822:0.014712,MtLYK2:0.05186283)1000:0.05694231,LjNFR1a:0.07999007)832:0.0215714,((GmNFR1a:0.03699912,GmNFR1b:0.02880309)677:0.01953081,CacLYK1:0.08167788)998:0.0365518)1000:0.18063742)999:0.06770856,(((((PpLYK1:0.06109585,PpLYK2:0.07592721)933:0.02843438,(MdLYK1:0.0383793,MdLYK2:0.05889107)1000:0.04350185)924:0.02731021,FvLYK1:0.1453431)999:0.05725737,AtCERK1:0.49326411)294:0.00688996,CasLYK1:0.16961626)430:0.02167592)577:0.02895946,(PtLYK1:0.03931642,PtLYK2:0.06215206)1000:0.13921412)942:0.04476039,VvLYK1:0.14879144)999:0.03856292,(((((((CacLYK3:0.06016158,GmLYK3:0.04988316)998:0.04732511,LjLYS7:0.08887798)598:0.01845951,MtLYK8:0.17283636)1000:0.15093267,(CasLYK2:0.00284497,CasLYK3:0.00922375)1000:0.16983704)341:0.01703062,((PpLYK3:0.0532316,MdLYK3:0.11516319)1000:0.02510313,FvLYK2:0.1079844)1000:0.07250312)492:0.02813366,PtLYK3:0.17773083)999:0.06833724,(VvLYK2:0.09020761,VvLYK3:0.13588809)1000:0.0759631)999:0.03856292);'

class Node_test(unittest.TestCase):
    def setUp(self):
        self.tree= egglib.Tree(string=TREE_STRING)

    def tearDown(self):
        del self.tree

    def test_Node_T(self):
        node= self.tree.find_clade(['VvLYK2'], ancestral=True)
        self.assertIsInstance(node, egglib._tree.Node)
    
    def test_newick_T(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        self.assertEqual(node.newick(), '(PtLYK3:0.17773083,((FvLYK2:0.1079844,(PpLYK3:0.0532316,MdLYK3:0.11516319)1000:0.02510313)1000:0.07250312,((((CacLYK3:0.06016158,GmLYK3:0.04988316)998:0.04732511,LjLYS7:0.08887798)598:0.01845951,MtLYK8:0.17283636)1000:0.15093267,(CasLYK2:0.00284497,CasLYK3:0.00922375)1000:0.16983704)341:0.01703062)492:0.02813366)999;')

    def test_leaves_down_T(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        self.assertEqual(node.leaves_down(), ['PtLYK3', 'FvLYK2', 'PpLYK3', 'MdLYK3', 'CacLYK3', 'GmLYK3', 'LjLYS7', 'MtLYK8', 'CasLYK2', 'CasLYK3'])


    def test_leaves_up_T(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        self.assertEqual(node.leaves_up(), ['VvLYK2', 'VvLYK3', 'VvLYK1', 'PtLYK1', 'PtLYK2', 'CasLYK1', 'AtCERK1', 'FvLYK1', 'PpLYK1', 'PpLYK2', 'MdLYK1', 'MdLYK2', 'CacLYK1', 'GmNFR1a', 'GmNFR1b', 'LjNFR1a', 'MtLYK2', 'MtLYK3', 'PsSYM37', 'CecLYK1', 'LjNFR1b', 'MtLYK1', 'LjNFR1c', 'MtLYK6', 'GmLYK2', 'CacLYK4', 'MtLYK7', 'CecLYK2', 'CacLYK2', 'GmLYK2b', 'LjLYS6', 'MtLYK9'])

    def test_has_descendant_T(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        node_c = self.tree.find_clade(['LjLYS7'],ancestral=True)
        node_c1 = self.tree.find_clade(['GmNFR1a'],ancestral=True)
        self.assertTrue(node.has_descendant(node_c))
        self.assertFalse(node.has_descendant(node_c1))

    def test_is_parent_T(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        node_c = self.tree.find_clade(['LjLYS7'],ancestral=True)
        self.assertFalse(node.is_parent(node_c))

    def test_parent_T(self):
        node = self.tree.find_clade(['LjNFR1a'],ancestral=True)
        self.assertEqual(node.parent.newick(), '(LjNFR1a:0.07999007,(MtLYK2:0.05186283,(MtLYK3:0.07978386,PsSYM37:0.07493097)822:0.014712)1000:0.05694231)832;')

    def test_child_T(self):
        node = self.tree.find_clade(['LjNFR1a', 'PsSYM37'], ancestral=True)
        self.assertEqual(node.child(0).newick(), 'LjNFR1a;')

    def test_child_E(self):
        node = self.tree.find_clade(['LjNFR1a', 'PsSYM37'], ancestral=True)
        with self.assertRaises(IndexError):
            node.child(1000)

    def test_set_branch_to_T(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        A = node.branch_to(0)
        node.set_branch_to(0, 0.7734) 
        B = node.branch_to(0)
        self.assertEqual(B, 0.7734)
        self.assertNotEqual(A,B)

    def test_set_branch_to_E(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        with self.assertRaises(ValueError):
            node.set_branch_to(1500, 0.7734) 
        node_e = self.tree.find_clade(['LjNFR1a'],ancestral=True)
        with self.assertRaises(ValueError):
            node.set_branch_to(node_e, 0.7734)

    def test_branch_to_T(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        A = node.branch_to(0)
        self.assertEqual(A, 0.17773083)
        node1= node.child(1)
        B = node.branch_to(node1)
        self.assertEqual(B, 0.02813366)

    def test_branch_to_E(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        with self.assertRaises(ValueError):
            node.branch_to(1500)
        node_e = self.tree.find_clade(['LjNFR1a'],ancestral=True)
        with self.assertRaises(ValueError):
            node.branch_to(node_e)

    def test_parent_branch_getter_T(self):
        node = self.tree.find_clade(['LjNFR1a'],ancestral=True)
        self.assertEqual(node.parent_branch, 0.07999007)
    
    def test_parent_branch_getter_E(self):
        node = self.tree._base
        with self.assertRaises(ValueError):
            node.parent_branch

    def test_parent_branch_setter_T(self):
        node = self.tree.find_clade(['LjNFR1a'],ancestral=True)
        node.parent_branch=0.05
        self.assertEqual(node.parent_branch, 0.05)
    
    def test_parent_branch_setter_E(self):
        node = self.tree._base
        with self.assertRaises(ValueError):
            node.parent_branch=0.5
    
    def test_label_getter_T(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        self.assertEqual(node.label, 999)

    def test_label_setter_T(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        node.label= 'new_label'
        self.assertEqual(node.label, 'new_label')

    def test_num_children_T(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        self.assertEqual(node.num_children, 2)

class Tree_test(unittest.TestCase):
    def setUp(self):
        self.tree= egglib.Tree(string=TREE_STRING)

    def tearDown(self):
        del self.tree

    def test_Tree_T(self):
        self.assertIsInstance(self.tree, egglib.Tree)

    def test_copy_T(self):
        tree0=self.tree.copy()
        self.assertIsInstance(tree0, egglib.Tree)
        self.assertEqual(str(self.tree), str(tree0))

    def test_copy_E(self):
        node0 = self.tree.find_clade(['VvLYK2'], ancestral=True)
        with self.assertRaises(ValueError):
            tree0=self.tree.copy(node0)
        with self.assertRaises(ValueError):
            tree1=self.tree.copy(self.tree.find_clade(['error'], ancestral=True))
    
    def test_extract_T(self):
        node0 = self.tree.find_clade(names=['CacLYK1', 'MtLYK9'], ancestral=True)
        tree0=self.tree.extract(node0, 'clade-A')
        self.assertNotEqual(str(tree0), str(self.tree)) 

    def test_extract_E(self):
        with self.assertRaises(ValueError):
            tree0= self.tree.extract(self.tree.find_clade(['error']))
        with self.assertRaises(ValueError):
            node0 = self.tree.find_clade(['VvLYK3'], ancestral=True)
            tree0=self.tree.extract(node0)
        with self.assertRaises(ValueError):
            node0 = self.tree.base
            tree0=self.tree.extract(node0)

    def test__del__T(self):
        tree= egglib.Tree(string=TREE_STRING)
        del tree
        with self.assertRaises(NameError):
            tree

    def test__str__T(self):
        tree_s=str(self.tree)
        self.assertIsInstance(tree_s, str)
    
    def test_num_nodes_T(self):
        self.assertEqual(self.tree.num_nodes, 83)
    
    def test_num_leaves(self):
        self.assertEqual(self.tree.num_leaves, 42)
    
    def test_newick_T(self):
        n0=str(self.tree.newick())
        n1=str(self.tree.newick(skip_labels=True, skip_brlens=True))
        self.assertEqual(n1, '(((VvLYK2,VvLYK3),(PtLYK3,((FvLYK2,(PpLYK3,MdLYK3)),((((CacLYK3,GmLYK3),LjLYS7),MtLYK8),(CasLYK2,CasLYK3))))),(VvLYK1,((PtLYK1,PtLYK2),((CasLYK1,(AtCERK1,(FvLYK1,((PpLYK1,PpLYK2),(MdLYK1,MdLYK2))))),(((CacLYK1,(GmNFR1a,GmNFR1b)),(LjNFR1a,(MtLYK2,(MtLYK3,PsSYM37)))),((CecLYK1,(((LjNFR1b,MtLYK1),(LjNFR1c,MtLYK6)),((GmLYK2,CacLYK4),MtLYK7))),(CecLYK2,((CacLYK2,GmLYK2b),(LjLYS6,MtLYK9)))))))));')
        self.assertEqual(n0, str(self.tree))

    def test_base_T(self):
        node_b=self.tree.base
        self.assertIsInstance(node_b, egglib._tree.Node)

    def test_add_node_T(self):
        node = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        self.tree.add_node(node, 'new_node', 0.54)
        new_node = self.tree.find_clade(['new_node'], ancestral=True)
        self.assertIsNotNone(new_node)

    def test_add_node_E(self):
        node0 = self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        tree0 = self.tree.extract(node0)
        with self.assertRaises(ValueError):
            self.tree.add_node(node0, 'new_node' , 0.54)

    def test_iter_leaves_T(self):
        self.assertIsInstance(self.tree.iter_leaves(), collections.abc.Iterable)
        
    def test_get_leaf_T(self):
        leaf=self.tree.get_leaf('CasLYK1')
        self.assertIsInstance(leaf, egglib._tree.Node)
        self.assertEqual(str(leaf.newick()), 'CasLYK1;')

    def test_depth_iter_T(self):
        self.assertIsInstance(self.tree.depth_iter(), collections.abc.Iterable)
        node=self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        i=1
        for n in self.tree.depth_iter(node):
            self.assertIsInstance(n, egglib._tree.Node)
            i+=1
        self.assertEqual(i, 20)

    def test_depth_iter_E(self):
        node= self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        self.tree.extract(node)
        with self.assertRaises(ValueError):
            self.tree.depth_iter(node)

    def test_total_length_T(self):
        L=self.tree.total_length()
        self.assertEqual(L, 6.25127455)

    def test_total_length_E(self):
        tree= egglib.Tree(string='(((VvLYK2,VvLYK3),(PtLYK3,((FvLYK2,(PpLYK3,MdLYK3)),((((CacLYK3,GmLYK3),LjLYS7),MtLYK8),(CasLYK2,CasLYK3))))),(VvLYK1,((PtLYK1,PtLYK2),((CasLYK1,(AtCERK1,(FvLYK1,((PpLYK1,PpLYK2),(MdLYK1,MdLYK2))))),(((CacLYK1,(GmNFR1a,GmNFR1b)),(LjNFR1a,(MtLYK2,(MtLYK3,PsSYM37)))),((CecLYK1,(((LjNFR1b,MtLYK1),(LjNFR1c,MtLYK6)),((GmLYK2,CacLYK4),MtLYK7))),(CecLYK2,((CacLYK2,GmLYK2b),(LjLYS6,MtLYK9)))))))));')
        with self.assertRaises(ValueError):
            tree.total_length()
        
    def test_find_clade_T(self):
        node = self.tree.find_clade(names=['CacLYK1', 'MtLYK9'], ancestral=True)
        self.assertIsInstance(node, egglib._tree.Node)

    def test_fin_clade_E(self):
        with self.assertRaises(ValueError):
            self.tree.find_clade(names=['CacLYK1', 'MtLYK9'], ancestral=True, both_sides=True)
        with self.assertRaises(ValueError):
            self.tree.find_clade(names=[], ancestral=True)

        tree= egglib.Tree(string='(((VvLYK2,VvLYK3),(PtLYK3,((FvLYK2,(PpLYK3,MdLYK3)),((((CacLYK3,GmLYK3),LjLYS7),MtLYK8),(CasLYK2,CasLYK3))))));')
        with self.assertRaises(ValueError):
            tree.find_clade(names=['VvLYK2','VvLYK3','PtLYK3','FvLYK2','PpLYK3','MdLYK3','CacLYK3','GmLYK3','LjLYS7','MtLYK8','CasLYK2','CasLYK3','VvLYK1','PtLYK1','PtLYK2','CasLYK1','AtCERK1','FvLYK1','PpLYK1','PpLYK2','MdLYK1','MdLYK2','CacLYK1','GmNFR1a','GmNFR1b','LjNFR1a','MtLYK2','MtLYK3','PsSYM37','CecLYK1','LjNFR1b','MtLYK1','LjNFR1c','MtLYK6','GmLYK2','CacLYK4','MtLYK7','CecLYK2','CacLYK2','GmLYK2b','LjLYS6','MtLYK9'])

    def test_collapse_T(self):
        node = self.tree.find_clade(['MtLYK3', 'GmNFR1a'], ancestral=True)
        self.assertEqual(self.tree.num_nodes, 83)
        self.tree.collapse(node)
        self.assertEqual(self.tree.num_nodes, 82)

    def test_collapse_E(self):
        node_b = self.tree.find_clade(['VvLYK2', 'PsSYM37'], ancestral=True)
        node_t = self.tree.find_clade(['AtCERK1'])
        tree= egglib.Tree(string='(((VvLYK2,VvLYK3),(PtLYK3,((FvLYK2,(PpLYK3,MdLYK3)),((((CacLYK3,GmLYK3),LjLYS7),MtLYK8),(CasLYK2,CasLYK3))))),(VvLYK1,((PtLYK1,PtLYK2),((CasLYK1,(AtCERK1,(FvLYK1,((PpLYK1,PpLYK2),(MdLYK1,MdLYK2))))),(((CacLYK1,(GmNFR1a,GmNFR1b)),(LjNFR1a,(MtLYK2,(MtLYK3,PsSYM37)))),((CecLYK1,(((LjNFR1b,MtLYK1),(LjNFR1c,MtLYK6)),((GmLYK2,CacLYK4),MtLYK7))),(CecLYK2,((CacLYK2,GmLYK2b),(LjLYS6,MtLYK9)))))))));')
        node_l = tree.find_clade(['MtLYK9'])
        with self.assertRaises(ValueError):
            self.tree.collapse(node_b)
        with self.assertRaises(ValueError):
            self.tree.collapse(node_t)
        with self.assertRaises(ValueError):
            tree.collapse(node_l)
        with self.assertRaises(ValueError):
            tree.collapse(node_l)
        node_m= self.tree.find_clade(['PtLYK3', 'CasLYK3'], ancestral=True)
        self.tree.extract(node_m)
        with self.assertRaises(ValueError):
            tree.collapse(node_m)
        
    def test_root_T(self):
        self.tree.collapse(self.tree.base.child(0))
        A=self.tree.num_nodes
        node = self.tree.find_clade(['MtLYK3', 'GmNFR1a'], ancestral=True)
        self.tree.root(node, 0.2)
        B=self.tree.num_nodes
        self.assertNotEqual(A, B)
        self.assertEqual(A, 82)
        self.assertEqual(B, 83)

    def test_root_E(self):
        node_b = self.tree.find_clade(['VvLYK2', 'PsSYM37'], ancestral=True)
        node = self.tree.find_clade(['MtLYK3', 'GmNFR1a'], ancestral=True)
        node_t = self.tree.find_clade(['AtCERK1'])
        with self.assertRaises(ValueError):
            self.tree.root(node_b, 0.2, reoriente=True)
        with self.assertRaises(ValueError):
            self.tree.root(node_t, 0.2, reoriente=True)
        with self.assertRaises(ValueError):
            self.tree.root(node_t, 0.2, reoriente=False)
        with self.assertRaises(ValueError):
            self.tree.root(node, 0.2, reoriente=False)

        tree= egglib.Tree(string='(((VvLYK2,VvLYK3),(PtLYK3,((FvLYK2,(PpLYK3,MdLYK3)),((((CacLYK3,GmLYK3),LjLYS7),MtLYK8),(CasLYK2,CasLYK3))))),(VvLYK1,((PtLYK1,PtLYK2),((CasLYK1,(AtCERK1,(FvLYK1,((PpLYK1,PpLYK2),(MdLYK1,MdLYK2))))),(((CacLYK1,(GmNFR1a,GmNFR1b)),(LjNFR1a,(MtLYK2,(MtLYK3,PsSYM37)))),((CecLYK1,(((LjNFR1b,MtLYK1),(LjNFR1c,MtLYK6)),((GmLYK2,CacLYK4),MtLYK7))),(CecLYK2,((CacLYK2,GmLYK2b),(LjLYS6,MtLYK9)))))))));')
        node_n =tree.find_clade(['MtLYK3', 'GmNFR1a'], ancestral=True)
        with self.assertRaises(ValueError):
            tree.root(node_n, 0.2, reoriente=True)

    def test_map_descendants_T(self):
        for node, labels in self.tree.map_descendants().items():
            self.assertIsInstance(node, egglib._tree.Node)
            self.assertIsInstance(labels, tuple)
    
    def test_frequency_node_T(self):
        tree0=str(self.tree)
        copies = [self.tree.copy() for i in range(1000)]
        for copy in copies:
            leaves = [i.label for i in copy.iter_leaves()]
            random.shuffle(leaves)
            for node, leaf in zip(copy.iter_leaves(), leaves): node.label = leaf
        self.tree.frequency_nodes(copies)
        tree1=str(self.tree)
        self.tree.frequency_nodes(copies, relative=True)
        tree2=str(self.tree)
        self.assertNotEqual(tree0, tree1)
        self.assertNotEqual(tree1, tree2)

    def test_frequency_node_E(self):
        tree_e=egglib.Tree(string='(Spider:0.01257025,Woolly:0.02023601,(Howler:0.03625789,((Titi:0.02002846,Saki:0.02646824):0.01312676,((Owl:0.02467454,(((Gorilla:0.00570008,(Human:0.00467442,Chimp:0.00218595):0.00198277):0.00803251,(Gibbon:0.02031871,Orangutan:0.01428695):0.00052501):0.01497214,(Colobus:0.00134948,(DLangur:0.00479108,(Patas:0.01038680,((AGM_cDNA:0.00067736,Tant_cDNA:0.00000006):0.00511480,(Baboon:0.00531400,Rhes_cDNA:0.00519922):0.00433750):0.00200495):0.00621549):0.00134574):0.02864316):0.11226732):0.00500924,(Squirrel:0.04657578,(PMarmoset:0.02254897,Tamarin:0.01990484):0.01687811):0.00000008):0.00119894):0.01464801):0.01145931);')
        tree_d=egglib.Tree(string='(((VvLYK2:0.09020761,VvLYK3:0.13588809)1000:0.0759631,(PtLYK3:0.17773083,((FvLYK2:0.1079844,(PpLYK3:0.0532316,MdLYK3:0.11516319)1000:0.02510313)1000:0.07250312,((((CacLYK3:0.06016158,GmLYK3:0.04988316)998:0.04732511,VvLYK2:0.08887798)598:0.01845951,MtLYK8:0.17283636)1000:0.15093267,(CasLYK2:0.00284497,CasLYK3:0.00922375)1000:0.16983704)341:0.01703062)492:0.02813366)999:0.06833724)999:0.03856292,(VvLYK1:0.14879144,((PtLYK1:0.03931642,PtLYK2:0.06215206)1000:0.13921412,((CasLYK1:0.16961626,(AtCERK1:0.49326411,(VvLYK2:0.1453431,((PpLYK1:0.06109585,VvLYK2:0.07592721)933:0.02843438,(MdLYK1:0.0383793,MdLYK2:0.05889107)1000:0.04350185)924:0.02731021)999:0.05725737)294:0.00688996)430:0.02167592,(((CacLYK1:0.08167788,(VvLYK2:0.03699912,GmNFR1b:0.02880309)677:0.01953081)998:0.0365518,(LjNFR1a:0.07999007,(MtLYK2:0.05186283,(MtLYK3:0.07978386,PsSYM37:0.07493097)822:0.014712)1000:0.05694231)832:0.0215714)1000:0.18063742,((CecLYK1:0.1680513,(((LjNFR1b:0.13030045,VvLYK2:0.15554318)1000:0.07303159,(LjNFR1c:0.13766267,MtLYK6:0.16025718)982:0.03580008)990:0.02671666,((GmLYK2:0.04933436,CacLYK4:0.05585385)1000:0.06029768,MtLYK7:0.1292539)997:0.04581387)1000:0.05564467)998:0.03954658,(CecLYK2:0.07693512,((CacLYK2:0.07392427,GmLYK2b:0.05439578)1000:0.05710725,(LjLYS6:0.06218823,MtLYK9:0.12383649)905:0.02507435)1000:0.08050891)794:0.01706889)820:0.01986008)999:0.06770856)577:0.02895946)942:0.04476039)999:0.03856292);')
        copies_e = [tree_e.copy() for i in range(1000)]
        for copy in copies_e:
            leaves = [i.label for i in copy.iter_leaves()]
            random.shuffle(leaves)
            for node, leaf in zip(copy.iter_leaves(), leaves): node.label = leaf
        with self.assertRaises(ValueError):
            self.tree.frequency_nodes(copies_e)

        copies_d = [tree_d.copy() for i in range(1000)]
        for copy in copies_d:
            leaves = [i.label for i in copy.iter_leaves()]
            random.shuffle(leaves)
            for node, leaf in zip(copy.iter_leaves(), leaves): node.label = leaf
        with self.assertRaises(ValueError):
            self.tree.frequency_nodes(copies_d)

        copies_n = []
        with self.assertRaises(ValueError):
            self.tree.frequency_nodes(copies_n, relative=True)

    def test_clean_internal_labels_T(self):
        self.tree.clean_internal_labels()
        self.assertEqual(str(self.tree), "(((VvLYK2:0.09020761,VvLYK3:0.13588809):0.0759631,(PtLYK3:0.17773083,((FvLYK2:0.1079844,(PpLYK3:0.0532316,MdLYK3:0.11516319):0.02510313):0.07250312,((((CacLYK3:0.06016158,GmLYK3:0.04988316):0.04732511,LjLYS7:0.08887798):0.01845951,MtLYK8:0.17283636):0.15093267,(CasLYK2:0.00284497,CasLYK3:0.00922375):0.16983704):0.01703062):0.02813366):0.06833724):0.03856292,(VvLYK1:0.14879144,((PtLYK1:0.03931642,PtLYK2:0.06215206):0.13921412,((CasLYK1:0.16961626,(AtCERK1:0.49326411,(FvLYK1:0.1453431,((PpLYK1:0.06109585,PpLYK2:0.07592721):0.02843438,(MdLYK1:0.0383793,MdLYK2:0.05889107):0.04350185):0.02731021):0.05725737):0.00688996):0.02167592,(((CacLYK1:0.08167788,(GmNFR1a:0.03699912,GmNFR1b:0.02880309):0.01953081):0.0365518,(LjNFR1a:0.07999007,(MtLYK2:0.05186283,(MtLYK3:0.07978386,PsSYM37:0.07493097):0.014712):0.05694231):0.0215714):0.18063742,((CecLYK1:0.1680513,(((LjNFR1b:0.13030045,MtLYK1:0.15554318):0.07303159,(LjNFR1c:0.13766267,MtLYK6:0.16025718):0.03580008):0.02671666,((GmLYK2:0.04933436,CacLYK4:0.05585385):0.06029768,MtLYK7:0.1292539):0.04581387):0.05564467):0.03954658,(CecLYK2:0.07693512,((CacLYK2:0.07392427,GmLYK2b:0.05439578):0.05710725,(LjLYS6:0.06218823,MtLYK9:0.12383649):0.02507435):0.08050891):0.01706889):0.01986008):0.06770856):0.02895946):0.04476039):0.03856292);"
)

    def test_clean_branch_lengths_T(self):
        self.tree.clean_branch_lengths()
        with self.assertRaises(ValueError):
            self.tree.total_length()

    def test_remove_node_T(self):
        node = self.tree.find_clade(['VvLYK2', 'CasLYK3'], ancestral=True)
        self.tree.remove_node(node)
        with self.assertRaises(ValueError):
            self.tree.find_clade(['VvLMK2', 'CasLYK3'], ancestral=True)

    def test_remove_node_E(self):
        node_b = self.tree.find_clade(['VvLYK2', 'PsSYM37'], ancestral=True)
        node = self.tree.find_clade(['VvLYK2', 'CasLYK3'], ancestral=True)
        with self.assertRaises(ValueError):
            self.tree.remove_node(node_b)
        self.tree.remove_node(node)
        with self.assertRaises(ValueError):
            self.tree.remove_node(node)

    def test_lateralize_T(self):
        self.tree.lateralize(reverse=True)
        self.assertEqual(str(self.tree),TREE_STRING_R) 

    def test_midroot_T(self):
        tree=egglib.Tree(string='(Spider:0.01257025,Woolly:0.02023601,(Howler:0.03625789,((Titi:0.02002846,Saki:0.02646824):0.01312676,((Owl:0.02467454,(((Gorilla:0.00570008,(Human:0.00467442,Chimp:0.00218595):0.00198277):0.00803251,(Gibbon:0.02031871,Orangutan:0.01428695):0.00052501):0.01497214,(Colobus:0.00134948,(DLangur:0.00479108,(Patas:0.01038680,((AGM_cDNA:0.00067736,Tant_cDNA:0.00000006):0.00511480,(Baboon:0.00531400,Rhes_cDNA:0.00519922):0.00433750):0.00200495):0.00621549):0.00134574):0.02864316):0.11226732):0.00500924,(Squirrel:0.04657578,(PMarmoset:0.02254897,Tamarin:0.01990484):0.01687811):0.00000008):0.00119894):0.01464801):0.01145931);')
        self.assertEqual(tree.num_nodes, 40)
        tree.midroot()
        self.assertEqual(tree.num_nodes, 41)

    def test_midroot_E(self):
        with self.assertRaises(ValueError):
            self.tree.midroot()
        tree_l=egglib.Tree(string='(Spider,Woolly,(Howler,((Titi,Saki),((Owl,(((Gorilla,(Human,Chimp)),(Gibbon,Orangutan)),(Colobus,(DLangur,(Patas,((AGM_cDNA,Tant_cDNA),(Baboon,Rhes_cDNA))))))),(Squirrel,(PMarmoset,Tamarin))))));')
        with self.assertRaises(ValueError):
            tree_l.midroot()
    
        tree_n=egglib.Tree(string='(Spider:0.01257025,Woolly:0.02023601);')
        with self.assertRaises(ValueError):
            tree_n.midroot()

    def test_unroot_T(self):
        tree=egglib.Tree(string='(Spider:0.01257025,Woolly:0.02023601,(Howler:0.03625789,((Titi:0.02002846,Saki:0.02646824):0.01312676,((Owl:0.02467454,(((Gorilla:0.00570008,(Human:0.00467442,Chimp:0.00218595):0.00198277):0.00803251,(Gibbon:0.02031871,Orangutan:0.01428695):0.00052501):0.01497214,(Colobus:0.00134948,(DLangur:0.00479108,(Patas:0.01038680,((AGM_cDNA:0.00067736,Tant_cDNA:0.00000006):0.00511480,(Baboon:0.00531400,Rhes_cDNA:0.00519922):0.00433750):0.00200495):0.00621549):0.00134574):0.02864316):0.11226732):0.00500924,(Squirrel:0.04657578,(PMarmoset:0.02254897,Tamarin:0.01990484):0.01687811):0.00000008):0.00119894):0.01464801):0.01145931);')
        self.assertEqual(tree.num_nodes, 40)
        tree.midroot()
        self.assertEqual(tree.num_nodes, 41)
        tree.unroot()
        self.assertEqual(tree.num_nodes, 40)

    def test_unroot_E(self):
        tree=egglib.Tree(string='(Spider:0.01257025,Woolly:0.02023601,(Howler:0.03625789,((Titi:0.02002846,Saki:0.02646824):0.01312676,((Owl:0.02467454,(((Gorilla:0.00570008,(Human:0.00467442,Chimp:0.00218595):0.00198277):0.00803251,(Gibbon:0.02031871,Orangutan:0.01428695):0.00052501):0.01497214,(Colobus:0.00134948,(DLangur:0.00479108,(Patas:0.01038680,((AGM_cDNA:0.00067736,Tant_cDNA:0.00000006):0.00511480,(Baboon:0.00531400,Rhes_cDNA:0.00519922):0.00433750):0.00200495):0.00621549):0.00134574):0.02864316):0.11226732):0.00500924,(Squirrel:0.04657578,(PMarmoset:0.02254897,Tamarin:0.01990484):0.01687811):0.00000008):0.00119894):0.01464801):0.01145931);')
        with self.assertRaises(ValueError):
            tree.unroot()
        
        tree=egglib.Tree(string='((((Gorilla:0.00570008,(Human:0.00467442,Chimp:0.00218595):0.00198277):0.00803251,(Gibbon:0.02031871,Orangutan:0.01428695):0.00052501):0.01497214,(Colobus:0.00134948,(DLangur:0.00479108,(Patas:0.0103868,((AGM_cDNA:0.00067736,Tant_cDNA:6e-08):0.0051148,(Baboon:0.005314,Rhes_cDNA:0.00519922):0.0043375):0.00200495):0.00621549):0.00134574):0.02864316),(Owl:0.02467454,((Squirrel:0.04657578,(PMarmoset:0.02254897,Tamarin:0.01990484):0.01687811):8e-08,((Titi:0.02002846,Saki:0.02646824):0.01312676,(Howler:0.03625789,(Spider:0.01257025,Woolly:0.02023601):0.01145931):0.01464801):0.00119894):0.00500924):0.05150704);')
        with self.assertRaises(ValueError):
            tree.unroot()



import subprocess, sys, os, shutil, platform

# list of python 3 sub-versions to compile for
minor_versions = 7, 8, 9, 10

# executable name on windows/posix
executable = os.path.join('Scripts' if platform.system()=='Windows' else 'bin', 'python')

# function to run a command
def run(args):
    ret = subprocess.run(args, shell=False)
    if ret.returncode != 0:
        sys.exit('an error occurred')

# get the script directory and move to EggLib to directory
path = os.path.dirname(os.path.abspath(__file__))
os.chdir(os.path.join(path, '..'))

# run for all python versions
for v in minor_versions:
    version = f'3.{v}'
    print(f'### python {version} ###')

    # precleaning
    if os.path.isdir(f'venv-{version}'):
        shutil.rmtree(f'venv-{version}')

    if platform.system()=='Windows': cmd = ['py', f'-{version}']
    else: cmd = [f'python{version}']

    # compilation
    try:
        print('creating virtual environment')
        run(cmd + ['-m', 'venv', f'venv-{version}'])

        print('setting up environment')
        run([os.path.join(f'venv-{version}', executable), '-m', 'pip', 'install', '--upgrade', 'pip', 'wheel'])

        print('compiling')
        run([os.path.join(f'venv-{version}', executable), 'setup.py', 'build', 'bdist_wheel'])

    # clean environment directory
    finally:
        if os.path.isdir(f'venv-{version}'):
            shutil.rmtree(f'venv-{version}')

    # testing
    startpath = os.getcwd()
    try:
        print('creating virtual environment')
        run(cmd + ['-m', 'venv', f'venv-{version}-testing'])
        os.chdir(f'venv-{version}-testing')

        print('setting up environment')
        run([executable, '-m', 'pip', 'install', '--upgrade', 'pip', 'wheel', 'scipy'])

        print('installing')
        run([executable, '-m', 'pip', 'install', 'egglib', '--no-index', '--find-links', os.path.join('..', 'dist')])

        print('testing')
        run([executable, os.path.join('..', 'test'), '-bictso', os.path.join('..', f'test-output-{version}.txt')])

    # clean environment directory
    finally:
        os.chdir(startpath)
        if os.path.isdir(f'venv-{version}-testing'):
            shutil.rmtree(f'venv-{version}-testing')



"""
    Copyright 2023 Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from .. import __version__, config, wrappers
import click, pathlib

@click.group()
def main():
    pass

#### command for editing app paths from the console #####################

@main.command(no_args_is_help=True)

@click.option('-z', '--zero', nargs=1, multiple=True,
    help='Format: --zero APP. Reset application configuration, '
         'dismissing previous configuration.')

@click.option('-Z', '--zero-all', is_flag=True, default=False,
    help='Reset all applications configuration, dismissing all '
         'previous configuration. Doesn\'t clear the persistent '
         'configuration file unless --save is used.')

@click.option('-a', '--auto', is_flag=True, default=False,
    help='Try default command names, which are usually the application \
name. This will work if the external applications are installed in the \
executable path. Only applications that are not currently configured \
are considered. If the test doesn\'t succeed, the error message is \
printed and the application is not configured. The default for \
individual applications can be overriden by --force-cmd, --cmd and \
--path options.')

@click.option('-c', '--cmd', 'commands', nargs=2, multiple=True,
    help='Format: --cmd APP COMMAND. Set the command used to start the \
application. To specify a full or relative path to the executable, use \
--path. Overrides the defaults specified by --auto. The is no priority \
between --force-cmd, --cmd and --path (the latest specified prevails). \
For --cmd and --path, if the test doesn\'t succeed, an error is \
caused.')

@click.option('-f', '--force-cmd', 'fcommands', nargs=2, multiple=True,
    help='Format: --force-cmd APP COMMAND. Set an application command. \
Like --cmd, but disable testing. The user needs to ensure that the \
software will be available through this command at runtime.')

@click.option('-p', '--path', 'paths', nargs=2, multiple=True,
    help='Format: --path APP PATH. Set the relative or absolute path to \
start the application. Relative paths are reformatted to be absolute. \
Overrides the defaults specified by --auto. The is no priority between \
--force-cmd, --cmd and --path (the latest specified prevails). For \
--cmd and--path, ff the test doesn\'t succeed, an error is caused.',
   type=(str, click.Path(exists=True, dir_okay=False, executable=True,
         resolve_path=True, path_type=pathlib.Path)))

@click.option('-L', '--list', 'list_', is_flag=True, default=False,
    help='Display application configuration. If any of --auto, --cmd, \
and --path is used, the final result is displayed.')

@click.option('-s', '--save', default=False, is_flag=True,
    help='Save configuration, after any change mandated by --clear \
--auto, --cmd and/or --path, in persistent file. The location of the \
file is controlled by --global and --user options.')

@click.option('-g', '--global', 'dest', flag_value='global',
    help='Save configuration in EggLib installation. This will affect \
all users using this installation of EggLib and might require \
administrative rights (like the EggLib installation).  The default is \
to use the current configuration file.')

@click.option('-u', '--user', 'dest', flag_value='user',
    help='Save configuration in user settings. This guarantees that \
only the current user will see the results when importing EggLib. \
Create the user-specific configuration file if it does not exist. The \
default is to use the current configuration file.')

@click.option('--delete-user', default=False, is_flag=True,
    help='Delete user configuration file, if any. This operation is \
not affected by any configuration operation specified by --clear, \
--auto, --cmd, and --path options. It is technically possible to \
save the configuration in the user\'s settings and then delete it \
using this function. Note that this operation is always performed \
last. To ignore user configuration and regenerate it, use --zero.')

def apps(zero_all, zero, auto, commands, fcommands, paths, list_, save, dest, delete_user):
    """
    Manage external application command names or paths. These are the
    commands who are used whenever EggLib uses an external application.
    Actions triggers by options are performed in the order in which they
    are listed.
    """

    if zero_all:
        wrappers.paths.clear()

    for key in zero:
        if not (app := wrappers.paths.get(key)):
            raise click.ClickException(f'Unknown application: "{key}"')
        app.zero()

    if auto:
        wrappers.paths.autodetect(True)

    for key, cmd in commands:
        if not (app := wrappers.paths.get(key)):
            raise click.ClickException(f'Unknown application: "{key}"')
        if (error := app.set_path(cmd, False)) != None:
            raise click.ClickException(f'Invalid command for {key}: "{cmd}"; error was:\n{error}')

    for key, cmd in fcommands:
        if not (app := wrappers.paths.get(key)):
            raise click.ClickException(f'Unknown application: "{key}"')
        app.set_path_force(cmd)

    for key, path in paths:
        if not (app := wrappers.paths.get(key)):
            raise click.ClickException(f'Unknown application: "{key}"')
        if (error := app.set_path(path, False)) != None:
            raise click.ClickException(f'Invalid command for {key}: "{path}"; error was:\n{error}')

    if list_:
        click.echo('Current values of application paths:')
        for p in wrappers.paths:
            if wrappers.paths[p] is None: click.echo(f'    - {p} (none)')
            else: click.echo(f'    + {p}: `{wrappers.paths[p]}\'')

    if save:
        wrappers.paths.save(dest)
        click.echo(f'Application paths saved to {wrappers.paths.fname}')

    if delete_user:
        wrappers.paths.delete_user()

#### just display version ##############################################

@main.command()
def version():
    click.echo(__version__)

#### display more information ##########################################

@main.command()
def infos():
    """
    Display information on EggLib installation.
    """

    click.echo(f'EggLib version {__version__}')
    click.echo(f'Installation path: {wrappers.__path__[0]}')
    click.echo(f'External application configuration file: {wrappers.paths.fname}')
    click.echo(f'debug flag: {config.debug:d}')
    click.echo(f'htslib flag: {config.htslib:d}')
    click.echo(f'version of muscle: {wrappers.paths.get("muscle").config["version"]}')



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import os, egglib, unittest, pathlib, tempfile
import collections
path = pathlib.Path(__file__).parent / '..' / 'data'

#GLOBAL VARIABLES FOR THE CLASS GENBANK:
positions=("1..1374,1528..2664,2756..3637,3647..3850,3948..4313,4381..5514,5672..6241,9785..10057,"
       "10535..11815,11816..13084,13092..13634,13656..15629,15967..16437,22900..23736,24287..25582,"
       "25685..26653,26740..27918,30050..30757,30761..34489,34492..35946,35987..37024,37582..39129,"
       "39255..40517,40543..41031,41180..42097,42141..42902,42975..44201,44203..45495,45902..46066,"
       "46411..47160,47170..47748,49838..51550,51652..51825,52120..53121,53869..54504,54736..55650,"
       "56055..57089,58268..59494,59519..59632,59685..61208,63302..63892,64533..65186,65187..66404,"
       "66456..67607,67843..68241,68341..68607,69038..69358")

positions_F=("1..1374,1528..2664,2756..3637,3647..3850,3948..4313,4381..5514,5672..6241,6241..9735,9785..10057,"
             "10044..10412,10409..10522,10535..11815,11816..13084,13092..13634,13656..15629,15967..16437,"
         "22900..23736,23726..24241,24287..25582,25685..26653,26740..27918,27905..28687,28684..29691,"
         "29684..29932,30050..30757,30761..34489,34492..35946,35987..37024,37021..37572,37582..39129,"
         "39255..40517,40543..41031,41180..42097,42141..42902,42975..44201,44203..45495,45902..46066,"
         "46411..47160,47170..47748,47709..48191,48184..49056,49056..49835,49838..51550,51652..51825,"
         "52120..53121,53121..53867,53869..54504,54736..55650,56055..57089,58268..59494,59519..59632,"
         "59685..61208,61205..61342,61323..63263,63302..63892,64533..65186,65187..66404,66456..67607,"
         "67843..68241,68341..68607,68607..69026,69038..69358")

position_MtChr2=("3644..5878,5950..5958,6182..6221,7110..7501,11376..11978,12092..12169,12561..12804,"
                "13200..13290,13753..13891,14477..14899,14987..15184,16128..16239,16346..16458,18831..18931,"
                "19048..19108,19221..19493,19881..19943,20032..20131,20443..20525,21480..21533,22105..22182,"
                "22276..22332,22407..22622,23909..24046,24214..24384,24917..25129,25566..26114,26840..26983,"
                "38765..39166,39230..39696,39903..40111,40332..40894,40955..41281,49014..49415,49513..49581,"
                "49668..50276,50445..50627,50722..50781,50882..51085,51246..51494,51710..51808,52984..53178,"
                "53552..53644,53752..54018,54215..54433,57408..57544,57959..57986,60428..61024,61175..61257,"
                "62081..62445,62872..62998,70784..70795,71844..72154,72647..72743,72826..72903,73553..74269,"
                "75534..75899,77481..78245,78305..78451,79456..79659,79770..80000,83525..83683,85767..86033,"
                "86109..86216,86314..86866,86961..87350,88883..88899,89974..90135,90383..90461,90651..90775,"
                "92092..92424,94554..94627,94699..94762,95107..95254,95459..95577,96925..97192,97270..97706,"
                "98016..98132,98231..98527")

position_MtChr2E=("3644..5878,5950..5958,6182..6221,7110..7501,11376..11978,12092..12169,12561..12804,"
                "13200..13290,13753..13891,14477..14899,14987..15184,16128..16239,16346..16458,18831..18931,"
                "19048..19108,19221..19493,19881..19943,20032..20131,20443..20525,21480..21533,22105..22182,"
                "22276..22332,22407..22622,23909..24046,24214..24384,24917..25129,25566..26114,26840..26983,"
                "38765..39166,39230..39696,39903..40111,40332..40894,40955..41281,49014..49415,49513..49581,"
                "49668..50276,50445..50627,50722..50781,50882..51085,51246..51494,51710..51808,52984..53178,"
                "53552..53644,53752..54018,54215..54433,57408..57544,57959..57986,60428..61024,61175..61257,"
                "62081..62445,62872..62998,70784..70795,71844..72154,72647..72743,72826..72903,73553..74269,"
                "75534..75899,77481..78245,78305..78451,79456..79659,79770..80000,83525..83683,85767..86033,"
                "86109..86216,86314..86866,86961..87350,88883..88899,89974..90135,90383..90461,90651..90775,"
                "92092..92424,94554..94627,94699..94762,95107..95254,95459..95577,96925..97192,97270..97706,"
                "98016..98132,98231..98527,110000..110504") #list of locations with an error: 110000..110504

def total_pb(positions):
    """
    This method indicate the number of bases covered by all segments

        :param positions: a string whicht contains all the positions
        """
    total=0
    list_pos=[]
    list_pos=positions.split(',')
    for i in list_pos:
        start,end=i.split('..')
        total += (int(end)-int(start))+1
    return total

class GenBankFeatureLocation_test(unittest.TestCase):

    def setUp(self):
        global positions
        self.gbfl= egglib.io.GenBankFeatureLocation(positions)

    def tearDown(self):
        del self.gbfl

    def test_GenbankFeature_T(self):
        self.assertIsInstance(self.gbfl, egglib.io._genbank.GenBankFeatureLocation)


    def test__len__T(self):
        global positions
        self.assertEqual(len(self.gbfl), 47)


    def test__getitem__T(self):
        self.assertEqual(self.gbfl[2], (2755, 3636))
    
    
    def test__str__T(self):
        self.assertIsInstance(str(self.gbfl), str)


    def test__str__E(self):
        gbfl= egglib.io.GenBankFeatureLocation()
        with self.assertRaises(RuntimeError):
            str(gbfl)

    def test_copy_T(self):
        c=self.gbfl.copy()
        self.assertEqual(str(self.gbfl[35]), str(c[35]))
        self.assertEqual(len(self.gbfl), len(c))

    def test_set_comp_T(self):
        self.assertFalse(self.gbfl.is_complement())
        self.gbfl.set_complement()
        self.assertTrue(self.gbfl.is_complement())

    def test_set_forw_T(self):
        self.gbfl.set_complement()
        self.assertTrue(self.gbfl.is_complement())
        self.gbfl.set_forward()
        self.assertFalse(self.gbfl.is_complement())

    def test_is_comp_T(self):
        self.gbfl.set_complement()
        self.assertTrue(self.gbfl.is_complement())

    def test_as_ord_T(self):
        self.gbfl.as_order()
        self.assertFalse(self.gbfl.is_range())

    def test_as_ran_T(self):
        self.gbfl.as_range()
        self.assertTrue(self.gbfl.is_range())

    def test_is_ran_T(self):
        self.gbfl.as_range()
        self.assertTrue(self.gbfl.is_range())
        self.gbfl.as_order()
        self.assertFalse(self.gbfl.is_range())

    def test_shift_T(self):
        self.assertEqual(self.gbfl[0],(0, 1373))
        self.gbfl.shift(1)
        self.assertEqual(self.gbfl[0],[1, 1374])
        self.assertNotEqual(self.gbfl[0],[0, 1373]) #with shitf gbfl[i] return a list not a tuple

    def test_add_single_base_T(self):
        ns_a=len(self.gbfl)
        positions_n=69370
        self.gbfl.add_single_base(positions_n)
        ns_b=len(self.gbfl)
        self.assertTrue(ns_a<ns_b)

    def test_add_single_base_E(self):
        positions_n='69370'
        with self.assertRaises(TypeError):
            self.gbfl.add_single_base(positions_n)
        positions_n2=65000
        with self.assertRaises(ValueError):
            self.gbfl.add_single_base(positions_n2)

    def test_add_between_base_T(self):
        ns_a=len(self.gbfl)
        positions_n=69370
        self.gbfl.add_between_base(positions_n)
        ns_b=len(self.gbfl)
        self.assertTrue(ns_a<ns_b)

    def test_add_between_base_E(self):
        positions_n='69370'
        with self.assertRaises(TypeError):
            self.gbfl.add_between_base(positions_n)
        positions_n2=65000
        with self.assertRaises(ValueError):
            self.gbfl.add_between_base(positions_n2)

    def test_add_base_range_T(self):
        ns_a=len(self.gbfl)
        start=69670
        end=69850
        self.gbfl.add_base_range(start,end)
        ns_b=len(self.gbfl)
        self.assertTrue(ns_a<ns_b)

    def test_add_base_range_E(self):
        start='69370'
        end='69850'
        with self.assertRaises(TypeError):
            self.gbfl.add_base_range(start,end)
        start_2=65000
        end_2=67350
        with self.assertRaises(ValueError):
            self.gbfl.add_base_range(start_2, end_2)
        start_3=70000
        end_3=69950
        with self.assertRaises(ValueError):
            self.gbfl.add_base_range(start_3, end_3)

    def test_add_base_choice_T(self):
        ns_a=len(self.gbfl)
        start=69670
        end=69850
        self.gbfl.add_base_choice(start,end)
        ns_b=len(self.gbfl)
        self.assertTrue(ns_a<ns_b)

    def test_add_base_choice_E(self):
        start='69370'
        end='69850'
        with self.assertRaises(TypeError):
            self.gbfl.add_base_choice(start,end)
        start_2=65000
        end_2=67350
        with self.assertRaises(ValueError):
            self.gbfl.add_base_choice(start_2, end_2)
        start_3=70000
        end_3=69950
        with self.assertRaises(ValueError):
            self.gbfl.add_base_choice(start_3, end_3)
        start_4=69650
        end_4=69650
        with self.assertRaises(ValueError):
            self.gbfl.add_base_choice(start_4, end_4)

    def test_rc_T(self):
        ns_a=len(self.gbfl)
        l1_loc=self.gbfl[ns_a-1]
        end_1=l1_loc[1]
        l_seq=70000
        self.gbfl.rc(l_seq)
        f2_loc=self.gbfl[0]
        start_2=f2_loc[0]
        self.assertEqual(((l_seq-1)-end_1),start_2)
    
class GenbankFeature_test(unittest.TestCase):
    def setUp(self):
        global positions
        self.gbfl= egglib.io.GenBankFeatureLocation(positions)
        self.gbf= egglib.io.GenBankFeature(self.gbfl)

    def tearDown(self):
        del self.gbfl
        del self.gbf

    def test_GenbankFeature_T(self):
        self.assertIsInstance(self.gbf, egglib.io._genbank.GenBankFeature)

    def test_type_T(self):
        global positions
        gbfl= egglib.io.GenBankFeatureLocation(positions)
        gbf= egglib.io.GenBankFeature(gbfl)
        self.assertEqual(gbf.get_type(), '')

    def test_qualifiers_T(self):
        dict_q=self.gbf.qualifiers()
        self.assertEqual(len(dict_q), 0)

    def test_add_qualifiers_T(self):
        dict_q1=self.gbf.qualifiers()
        self.gbf.add_qualifier('note', 'This is an example of "note" for the add_qualifer test')
        dict_q2=self.gbf.qualifiers()
        self.assertTrue(len(dict_q1)< len(dict_q2))

    def test_update_T(self):
        p1=self.gbf._parent
        q1=self.gbf._qualifiers
        positions_n='join(11376..11978,12092..12169,12561..12804,13200..13290,13753..13891,14477..14899,14987..15184)'
        gbfl= egglib.io.GenBankFeatureLocation(positions_n)
        qualifiers={'gene':'AC140550_58','label':'Bromodomain', 'note':'This is an example of set for the set test'}

        self.gbf.update("CDS", gbfl, ** qualifiers)
        p2=self.gbf._parent
        q2=self.gbf._qualifiers
                
        self.assertEqual(p1,p2)
        self.assertTrue(len(q1)<len(q2))
        self.assertEqual(self.gbf._type, 'CDS')
        self.assertIsNotNone(self.gbf._location)

    def test_update_E(self):
        positions_n='join(11376..11978,12092..12169,12561..12804,13200..13290,13753..13891,14477..14899,14987..15184)'
        qualifiers={'gene':'AC140550_58','type':'Bromodomain', 'note':'This is an example of set for the set test'}
        gbfl= egglib.io.GenBankFeatureLocation(positions_n)
        with self.assertRaises(ValueError):
            self.gbf.update("CDS" , gbfl, ** qualifiers)

    def test_parse_T(self):
        information=("     CDS             join(6182..6221,7110..7501)                     \n"
                     "/gene='AC140550_57'                     \n"
                     "/label=Protein kinase"
                )
        self.gbf.parse(information)
        self.assertEqual(self.gbf._type, 'CDS')
        self.assertEqual(str(self.gbf._location),'join(6182..6221,7110..7501)')
        self.assertTupleEqual(self.gbf._qualifiers[1], ('label', 'Protein kinase'))

    def test_parse_E(self):
        information=("CDSjoin(6182..6221,7110..7501)/gene: 'AC140550_57'/label: Protein kinase")
        with self.assertRaises(IOError):
            self.gbf.parse(information)
    
        information=("     CDS             join(6182..6221,7110..7501)                     \n"
                     "gene='AC140550_57'                     \n"
                     "label=Protein kinase"
                )
        with self.assertRaises(IOError):
            self.gbf.parse(information)
            information=("     CDS             join(6182..6221,7110..7501)                     \n"
                     "/gene: 'AC140550_57'                     \n"
                     "/label: Protein kinase"
                )
        with self.assertRaises(IOError):
            self.gbf.parse(information)

    def test_get_sequence_T(self):
        fname_s='MtChr2_315000-415000.gb'
        gb= egglib.io.GenBank(os.path.join(path, fname_s))
        gbf= egglib.io.GenBankFeature(gb)
        global position_MtChr2

        n= egglib.io.GenBankFeatureLocation(position_MtChr2)
        gbf.update("CDS",n)

        self.assertEqual(len(gbf.get_sequence()),total_pb(position_MtChr2)) #len: counts all bases in the sequence returned

    def test_get_sequence_E(self):
        fname_s='MtChr2_315000-415000.gb'
        gb= egglib.io.GenBank(os.path.join(path, fname_s))
        gbf= egglib.io.GenBankFeature(gb)
        global position_MtChr2E

        n= egglib.io.GenBankFeatureLocation(position_MtChr2E)
        gbf.update("CDS",n)
        with self.assertRaises(RuntimeError):
            gbf.get_sequence()

    def test_start_T(self):
        self.gbf.update("CDS",self.gbfl)
        self.assertEqual(self.gbf.get_start(),0)

    def test_stop_T(self):
        self.gbf.update("CDS",self.gbfl)
        self.assertEqual(self.gbf.get_stop(),69357)

    def test_copy_T(self):
        fname_s='MtChr2_315000-415000.gb'
        gb= egglib.io.GenBank(os.path.join(path, fname_s))
        gbf= egglib.io.GenBankFeature(gb)
        global position_MtChr2
        n= egglib.io.GenBankFeatureLocation(position_MtChr2)
        gbf.update("CDS",n)
        new=gbf.copy(gb)
        t_gbfn=str(type(new))

        self.assertEqual(t_gbfn, "<class 'egglib.io._genbank.GenBankFeature'>")
        self.assertEqual(new.get_sequence(), gbf.get_sequence())

    def test_shift_T(self):
        self.gbf.update("CDS",self.gbfl)
        self.gbf.shift(100) #shifting on 100 bases
        self.assertEqual(self.gbf.get_start(),100)
        self.assertEqual(self.gbf.get_stop(),69457)

    def test_rc_T(self):
        fname_s='MtChr2_315000-415000.gb'
        gb= egglib.io.GenBank(os.path.join(path, fname_s))
        gbf= egglib.io.GenBankFeature(gb)
        global position_MtChr2
        gbfl= egglib.io.GenBankFeatureLocation(position_MtChr2)
        gbf.update("CDS",gbfl)

        total_b=(100001-1) #-1, because there is 100001 bases but in the class 'gbf', bases are counted starting at 0 not 1
        end1=gbf.get_stop()
        gbf.rc()
        start2=gbf.get_start()
        self.assertEqual((total_b-end1),start2)


class Genbank_test(unittest.TestCase):
    def setUp(self):
        fname_s='MtChr2_315000-415000.gb'
        self.gb= egglib.io.GenBank(os.path.join(path, fname_s))
        f, self.fname = tempfile.mkstemp()
        os.close(f)

    def tearDown(self):
        del self.gb
        if os.path.isfile(self.fname):
            os.remove(self.fname)

    def test_Genbank_T(self):
        self.assertIsInstance(self.gb, egglib.io._genbank.GenBank)

    def test_add_feature_T(self):
        global position_MtChr2
        gbfl= egglib.io.GenBankFeatureLocation(position_MtChr2)
        gbf= egglib.io.GenBankFeature(gbfl)
        n_feat1= len(self.gb._features)
        information=("     CDSnew             join(0000..0000,0000..0000)                     \n"
                     "/gene='IDK00000'                     \n"
                     "/label=Protein IDK000"
                )
        gbf.parse(information)
        self.gb.add_feature(gbf)
        n_feat2= len(self.gb._features)
        self.assertTrue(n_feat1<n_feat2)
    
    def test_n_of_features_T(self):
        self.assertEqual(self.gb.number_of_features(),83)

    def test_get_sequence_T(self):
        l_seq= len(self.gb.sequence)
        self.assertEqual(l_seq,100001)

    def test_set_sequence_T(self):
        l_seq1= len(self.gb.sequence)
        self.gb.sequence='CATGTTGTGTGTTGTTGTCATTGAGGTAACAATTGATTCATATCGTTTCTGTGAGTTTACTGTTGTTGGAACAGAAATTTGGAATTTGTGATTAAACGTT'
        l_seq2= len(self.gb.sequence)
        self.assertEqual(l_seq2,100)
        self.assertTrue(l_seq1>l_seq2)

    def test__iter__T(self):
        self.assertIsInstance(self.gb, collections.abc.Iterable)

    def test_extract_T(self):
        l_gb= len(self.gb)
        gb_ss=self.gb.extract(3000,16000)
        l_gb_ss= len(gb_ss)
        n_ft_gb=self.gb.number_of_features()
        n_ft_gbss=gb_ss.number_of_features()
        self.assertTrue(l_gb>l_gb_ss)
        self.assertTrue(n_ft_gb>n_ft_gbss)

    def test_extract_E(self):
        l_gb= len(self.gb)
        with self.assertRaises(ValueError):
            gb_ss=self.gb.extract(3000,160000)
        with self.assertRaises(ValueError):
            gb_ss=self.gb.extract(-3000,160000)

    def test_write_T(self):
        self.gb.write(self.fname)
        self.assertTrue(os.path.exists(self.fname))
        self.assertTrue(os.path.getsize(self.fname)>0)

    def test__str__T(self):
        self.assertIsInstance(str(self.gb), str)

    def test_write_stream_T(self):
        gb_ss=self.gb.extract(3000,16000)
        f = open(self.fname, 'w')
        gb_ss.write_stream(f) 
        self.assertTrue(os.path.exists(self.fname))
        self.assertTrue(os.path.getsize(self.fname)>0)
        f.close()

    def test_write_stream_E(self):
        with self.assertRaises(TypeError):
            self.gb.write_stream(self.fname) #erreur avec fname
        seq='CATGTTGTGTGTTGTTGTCATTGAGGTAACAATTGATTCATATCGTTTCTGTGAGTTTACTGTTGTTGGAACAGAAATTTGGAATTTGTGATTAAACGTT'
        sequence= seq*10000000
        self.gb.sequence=sequence
        f = open(self.fname, 'w')
        with self.assertRaises(IOError):
            self.gb.write_stream(f)
        f.close()

    def test_rc_T(self):
        sequence=self.gb.sequence
        total_b=(100001-1)
        end1=self.gb._features[(83-1)].get_stop()
        dict_f1=self.gb._features[0].qualifiers()
        g1=dict_f1['gene']
        self.gb.rc()

        sequence_rc=self.gb.sequence
        start2=self.gb._features[0].get_start()
        dict_f2=self.gb._features[(83-1)].qualifiers()

        g2=dict_f1['gene']
    
        self.assertNotEqual(sequence, sequence_rc)
        self.assertEqual((total_b-end1),start2)
        self.assertEqual(g1,g2)



